"""Modern API wrapper for Genshin Impact built on asyncio and pydantic.

Documentation: https://thesadru.github.io/genshin.py

Source Code: https://github.com/thesadru/genshin.py
"""
from . import models
from .client import *
from .constants import *
from .errors import *
from .utils import *
